/***************************************************************************
    copyright            : (C) 2005 by Brian Nickel
    email                : brian.nickel@gmail.com
    based on             : flacfile.cpp from TagLib
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

using System.Collections;
using System;

namespace TagLib.Flac
{
   public class File : TagLib.File
   {
      //////////////////////////////////////////////////////////////////////////
      // private properties
      //////////////////////////////////////////////////////////////////////////
      private Id3v2.FrameFactory id3v2_frame_factory;
      private Id3v2.Tag          id3v2_tag;
      private long               id3v2_location;
      private uint               id3v2_original_size;

      private Id3v1.Tag          id3v1_tag;
      private long               id3v1_location;

      private Ogg.XiphComment    comment;

      private Flac.Tag           tag;

      private Properties         properties;
      private ByteVector         stream_info_data;
      private ByteVector         xiph_comment_data;
      private long               flac_start;
      private long               stream_start;
      private long               stream_length;
      private bool               scanned;

      private bool               has_xiph_comment;
      private bool               has_id3v2;
      private bool               has_id3v1;
      
      
      //////////////////////////////////////////////////////////////////////////
      // public methods
      //////////////////////////////////////////////////////////////////////////
      public File (string file, Id3v2.FrameFactory frame_factory, bool read_properties, Properties.ReadStyle properties_style) : base (file)
      {
         id3v2_frame_factory = frame_factory;
         id3v2_tag           = null;
         id3v2_location      = -1;
         id3v2_original_size = 0;
         id3v1_tag           = null;
         id3v1_location      = -1;
         comment             = null;
         properties          = null;
         flac_start          = 0;
         stream_start        = 0;
         stream_length       = 0;
         scanned             = false;
         has_xiph_comment    = false;
         has_id3v2           = false;
         has_id3v1           = false;
         
         try {Mode = AccessMode.Read;}
         catch (Exception e) {return;}
         
         Read (read_properties, properties_style);
         
         Mode = AccessMode.Closed;
      }

      public File (string file, Id3v2.FrameFactory frame_factory, bool read_properties) : this (file, frame_factory, read_properties, Properties.ReadStyle.Average)
      {
      }

      public File (string file, Id3v2.FrameFactory frame_factory) : this (file, frame_factory, true)
      {
      }

      public File (string file, bool read_properties, Properties.ReadStyle properties_style) : this (file, Id3v2.FrameFactory.Instance, read_properties, properties_style)
      {
      }

      public File (string file, bool read_properties) : this (file, read_properties, Properties.ReadStyle.Average)
      {
      }

      public File (string file) : this (file, true)
      {
      }





      public override bool Save ()
      {
         if (IsReadOnly)
         {
            Debugger.Debug ("Flac.File.Save() - Cannot save to a read only file.");
            return false;
         }
         
         try {Mode = AccessMode.Write;}
         catch (Exception e)
         {
            Debugger.Debug ("Flac.File.Save() - Cannot save to a read only file.");
            return false;
         }

         // Create new vorbis comments
         GetXiphComment (true);

         xiph_comment_data = comment.Render (false);

         ByteVector v = ByteVector.FromUInt((uint) xiph_comment_data.Count);

         // Set the type of the comment to be a Xiph / Vorbis comment
         // (See scan() for comments on header-format)
         v [0] = 4;
         v.Add (xiph_comment_data);


         // If file already have comment => find and update it
         //                       if not => insert one
         // TODO: Search for padding and use that

         if (has_xiph_comment)
         {
            long next_page_offset = flac_start;
            Seek (next_page_offset);
            ByteVector header = ReadBlock (4);
            uint length = header.Mid(1, 3).ToUInt ();

            next_page_offset += length + 4;

            // Search through the remaining metadata

            byte block_type = (byte) (header[0] & 0x7f);
            bool last_block = (header[0] & 0x80) != 0;

            while (!last_block)
            {
               Seek (next_page_offset);

               header = ReadBlock (4);
               block_type = (byte) (header[0] & 0x7f);
               last_block = (header[0] & 0x80) != 0;
               length = header.Mid(1, 3).ToUInt ();

               // Type is vorbiscomment
               if(block_type == 4)
               {
                  v [0] = header [0];
                  Insert (v, next_page_offset, length + 4);
                  break;
               }

               next_page_offset += length + 4;
            }
         }
         else
         {
            long next_page_offset = flac_start;

            Seek (next_page_offset);

            ByteVector header = ReadBlock (4);
            bool last_block = (header[0] & 0x80) != 0;
            uint length = header.Mid(1, 3).ToUInt ();

            // If last block was last, make this one last

            if(last_block)
            {
               // Copy the bottom seven bits into the new value

               ByteVector h = (byte)(header[0] & 0x7F);
               Insert (h, next_page_offset, 1);

               // Set the last bit
               v [0] = (byte) (v [0] | 0x80);
            }

            Insert (v, next_page_offset + length + 4, 0);
            has_xiph_comment = true;
         }

         // Update ID3 tags

         if(id3v2_tag != null)
         {
            if (has_id3v2)
               Insert (id3v2_tag.Render (), id3v2_location, id3v2_original_size);
            else
               Insert (id3v2_tag.Render (), 0, 0);
         }

         if(id3v1_tag != null)
         {
            if(has_id3v1)
               Seek (-128, System.IO.SeekOrigin.End);
            else
               Seek (0, System.IO.SeekOrigin.End);
            WriteBlock (id3v1_tag.Render ());
         }
         
         Mode = AccessMode.Closed;
         return true;
      }

      public Id3v2.Tag GetId3v2Tag (bool create)
      {
         if (create && id3v2_tag == null)
         {
            id3v2_tag = new Id3v2.Tag ();
            if (tag != null)
               TagLib.Tag.Duplicate (tag, id3v2_tag, true);
            tag = new Flac.Tag (comment, id3v2_tag, id3v1_tag);
         }
         return id3v2_tag;
      }
      
      public Id3v1.Tag GetId3v1Tag (bool create)
      {
         if (create && id3v1_tag == null)
         {
            id3v1_tag = new Id3v1.Tag ();
            if (tag != null)
               TagLib.Tag.Duplicate (tag, id3v1_tag, true);
            tag = new Flac.Tag (comment, id3v2_tag, id3v1_tag);
         }
         return id3v1_tag;
      }

      public Ogg.XiphComment GetXiphComment (bool create)
      {
         if (create && comment == null)
         {
            comment = new Ogg.XiphComment ();
            if (tag != null)
               TagLib.Tag.Duplicate (tag, comment, true);
            tag = new Flac.Tag (comment, id3v2_tag, id3v1_tag);
         }
         return comment;
      }

      public Id3v2.FrameFactory Id3v2FrameFactory
      {
         set {id3v2_frame_factory = value;}
      }
      
      
      //////////////////////////////////////////////////////////////////////////
      // public properties
      //////////////////////////////////////////////////////////////////////////
      public override TagLib.Tag Tag {get {return tag;}}

      public override AudioProperties AudioProperties {get {return properties;}}
      
      
      //////////////////////////////////////////////////////////////////////////
      // private methods
      //////////////////////////////////////////////////////////////////////////
      private void Read (bool read_properties, Properties.ReadStyle properties_style)
      {
         // Look for an ID3v2 tag
         id3v2_location = FindId3v2 ();

         if(id3v2_location >= 0) {

            id3v2_tag = new Id3v2.Tag (this, id3v2_location, id3v2_frame_factory);

            id3v2_original_size = id3v2_tag.Header.CompleteTagSize;

            if (id3v2_tag.Header.TagSize <= 0)
               id3v2_tag = null;
            else
               has_id3v2 = true;
         }

         // Look for an ID3v1 tag

         id3v1_location = FindId3v1 ();

         if(id3v1_location >= 0)
         {
            id3v1_tag = new Id3v1.Tag (this, id3v1_location);
            has_id3v1 = true;
         }

         // Look for FLAC metadata, including vorbis comments

         Scan ();

         if (!IsValid) return;

         if (has_xiph_comment)
            comment = new Ogg.XiphComment (XiphCommentData);

         if (has_xiph_comment || has_id3v2 || has_id3v1)
            tag = new Flac.Tag (comment, id3v2_tag, id3v1_tag);
         else
            tag = new Flac.Tag (new Ogg.XiphComment ());

         if (read_properties)
            properties = new Properties (stream_info_data, stream_length, properties_style);
      }

      private void Scan ()
      {
         // Scan the metadata pages

         if(scanned || !IsValid)
            return;

         long next_page_offset;
         long file_size = Length;

         if (has_id3v2)
            next_page_offset = Find ("fLaC", id3v2_location + id3v2_original_size);
         else
            next_page_offset = Find ("fLaC");

         if (next_page_offset < 0)
         {
            Debugger.Debug ("Flac.File.Scan () -- FLAC stream not found");
            SetValid (false);
            return;
         }

         next_page_offset += 4;
         flac_start = next_page_offset;

         Seek (next_page_offset);

         ByteVector header = ReadBlock (4);

         // Header format (from spec):
         // <1> Last-metadata-block flag
         // <7> BLOCK_TYPE
         //	0 : STREAMINFO
         //    1 : PADDING
         //    ..
         //    4 : VORBIS_COMMENT
         //    ..
         // <24> Length of metadata to follow

         byte block_type = (byte) (header [0] & 0x7f);
         bool last_block = (header [0] & 0x80) != 0;
         uint length = header.Mid (1, 3).ToUInt ();

         // First block should be the stream_info metadata
         if (block_type != 0)
         {
            Debugger.Debug ("Flac.File.Scan() -- invalid FLAC stream");
            SetValid (false);
            return;
         }
         stream_info_data  = ReadBlock ((int) length);
         next_page_offset += length + 4;

         // Search through the remaining metadata

         while (!last_block)
         {
            header = ReadBlock (4);
            block_type = (byte) (header [0] & 0x7f);
            last_block = (header [0] & 0x80) != 0;
            length = header.Mid (1, 3).ToUInt ();

            if (block_type == 1)
            {
               // Debugger.Debug ("Flac.File.Scan() -- Padding found");
            }
            // Found the vorbis-comment
            else if(block_type == 4)
            {
               xiph_comment_data  = ReadBlock ((int) length);
               has_xiph_comment = true;
            }

            next_page_offset += length + 4;
            if (next_page_offset >= file_size)
            {
               Debugger.Debug ("Flac.File.Scan() -- FLAC stream corrupted");
               SetValid (false);
               return;
            }
            Seek (next_page_offset);
         }

         // End of metadata, now comes the datastream
         stream_start = next_page_offset;
         stream_length = Length - stream_start;
         if (has_id3v1)
            stream_length -= 128;

         scanned = true;
      }

      private long FindId3v1 ()
      {
         if (IsValid)
         {
            Seek (-128, System.IO.SeekOrigin.End);
            long p = Tell;

            if (ReadBlock (3) == Id3v1.Tag.FileIdentifier)
            return p;
         }
         
         return -1;
      }

      private long FindId3v2 ()
      {
         if (IsValid)
         {
            Seek (0);

            if (ReadBlock (3) == Id3v2.Header.FileIdentifier)
               return 0;
         }
         
         return -1;
      }

      private ByteVector XiphCommentData
      {
         get
         {
            return (IsValid && has_xiph_comment) ? xiph_comment_data : new ByteVector ();
         }
      }
   }
}
