/***************************************************************************
    copyright            : (C) 2005 by Brian Nickel
    email                : brian.nickel@gmail.com
    based on             : id3v2tag.cpp from TagLib
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

using System.Collections;
using System;

namespace TagLib.Id3v2
{
   public class Tag : TagLib.Tag
   {
      //////////////////////////////////////////////////////////////////////////
      // private properties
      //////////////////////////////////////////////////////////////////////////
      private long           tag_offset;
      private FrameFactory   factory;

      private Header         header;
      private ExtendedHeader extended_header;
      private Footer         footer;

      private ArrayList      frame_list;
      
      
      //////////////////////////////////////////////////////////////////////////
      // public methods
      //////////////////////////////////////////////////////////////////////////
      public Tag () : base ()
      {
         tag_offset      = -1;
         factory         = FrameFactory.Instance;
         header          = new Header ();
         extended_header = null;
         footer          = null;
         frame_list      = new ArrayList ();
      }
      
      public Tag (File file, long tag_offset, FrameFactory factory) : this ()
      {
         this.tag_offset = tag_offset;
         this.factory = factory;
         
         Read (file);
      }

      public Frame [] GetFrames ()
      {
         return (Frame []) frame_list.ToArray (typeof (Frame));
      }

      public Frame [] GetFrames (ByteVector id)
      {
         ArrayList l = new ArrayList ();
         foreach (Frame f in frame_list)
            if (f.FrameId == id)
               l.Add (f);
         
         return (Frame []) l.ToArray (typeof (Frame));
      }

      public void AddFrame (Frame frame)
      {
        frame_list.Add (frame);
      }

      public void RemoveFrame (Frame frame)
      {
         frame_list.Remove (frame);
      }

      public void RemoveFrames (ByteVector id)
      {
         for (int i = frame_list.Count; i >= 0; i --)
         {
            Frame f = (Frame) frame_list [i];
            if (f.FrameId == id)
               RemoveFrame (f);
         }
      }

      public ByteVector Render ()
      {
         // We need to render the "tag data" first so that we have to correct size to
         // render in the tag's header.  The "tag data" -- everything that is included
         // in ID3v2::Header::tagSize() -- includes the extended header, frames and
         // padding, but does not include the tag's header or footer.

         ByteVector tag_data = new ByteVector ();

         // TODO: Render the extended header.
         header.ExtendedHeader = false;

         // Loop through the frames rendering them and adding them to the tagData.

         foreach (Frame frame in frame_list)
            if (!frame.Header.TagAlterPreservation)
               tag_data.Add (frame.Render ());

         // Compute the amount of padding, and append that to tagData.

         uint padding_size = 0;
         uint original_size = header.TagSize;

         if (!header.FooterPresent)
         {
            if (tag_data.Count < original_size)
               padding_size = (uint) (original_size - tag_data.Count);
            else
               padding_size = 1024;
            
            tag_data.Add (new ByteVector ((int) padding_size));
         }
         
         // Set the tag size.
         header.TagSize = (uint) tag_data.Count;

         if (header.FooterPresent)
         {
            footer = new Footer (header);
            return header.Render () + tag_data + footer.Render ();
         }
         
         return header.Render () + tag_data;
      }
      
      
      //////////////////////////////////////////////////////////////////////////
      // public properties
      //////////////////////////////////////////////////////////////////////////
      public override string Title
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TIT2"))
               return f.ToString ();
            return null;
         }
         set
         {
            SetTextFrame ("TIT2", value);
         }
      }
      
      public override string [] Artists
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TPE1"))
               return f.FieldList.ToArray ();
            return null;
         }
         set
         {
            SetTextFrame ("TPE1", new StringList (value));
         }
      }
      
      public override string [] Performers
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TPE2"))
               return f.FieldList.ToArray ();
            return null;
         }
         set
         {
            SetTextFrame ("TPE2", new StringList (value));
         }
      }
      
      public override string [] Composers
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TCOM"))
               return f.FieldList.ToArray ();
            return null;
         }
         set
         {
            SetTextFrame ("TCOM", new StringList (value));
         }
      }
      
      public override string Album
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TALB"))
               return f.ToString ();
            return null;
         }
         set
         {
            SetTextFrame ("TALB", value);
         }
      }
      
      public override string Comment
      {
         get
         {
            foreach (CommentsFrame f in GetFrames ("COMM"))
               return f.ToString ();
            return null;
         }
         set
         {
            if (value == null || value == "")
            {
               RemoveFrames ("COMM");
               return;
            }
            
            Frame [] frames = GetFrames ("COMM");
            if (frames.Length != 0)
               frames [0].SetText (value);
            else
            {
               CommentsFrame f = new CommentsFrame (factory.DefaultTextEncoding);
               AddFrame (f);
               f.SetText (value);
            }
         }
      }
      
      public override string [] Genres
      {
         get
         {
            ArrayList l = new ArrayList ();
            
            Frame [] frames = GetFrames ("TCON");
            TextIdentificationFrame frame;
            if (frames.Length != 0 && (frame = (TextIdentificationFrame) frames [0]) != null)
            {
               if(header.MajorVersion == 4)
               {
                  StringList fields = frame.FieldList;

                  foreach (string s in fields)
                  {
                     bool is_number = true;
                     foreach (char c in s)
                        if (c < '0' || c > '9')
                           is_number = false;

                     if(is_number)
                     {
                        try
                        {
                           l.Add (Id3v1.GenreList.Genre (Byte.Parse (s)));
                           continue;
                        }
                        catch (Exception e) {}
                     }
                     
                     l.Add (s);
                  }
               }
               else
               {
                  string s = frame.ToString ();

                  // ID3v2.3 "content type" can contain a ID3v1 genre number in parenthesis at
                  // the beginning of the field.  If this is all that the field contains, do a
                  // translation from that number to the name and return that.  If there is a
                  // string folloing the ID3v1 genre number, that is considered to be
                  // authoritative and we return that instead.  Or finally, the field may
                  // simply be free text, in which case we just return the value.

                  int closing = s.IndexOf (')');
                  if (s.Substring (0, 1) == "(" && closing > 0)
                  {
                     if(closing == s.Length - 1)
                     {
                        try
                        {
                           l.Add (Id3v1.GenreList.Genre (Byte.Parse (s.Substring (1, closing - 1))));
                        }
                        catch (System.Exception e)
                        {
                           l.Add (s);
                        }
                     }
                     else
                        l.Add (s.Substring (closing + 1));
                  }
                  else
                     l.Add (s);
               }
            }
            
            return (string []) l.ToArray (typeof (string));
         }
         set
         {
            for (int i = 0; i < value.Length; i ++)
            {
               int index = Id3v1.GenreList.GenreIndex (value [i]);
               if (index != 255)
                  value [i] = index.ToString ();
            }
            
            SetTextFrame ("TALB", new StringList (value));
         }
      }
      public override uint Year
      {
         get
         {
            foreach (TextIdentificationFrame f in GetFrames ("TDRC"))
            {
               try
               {
                  return UInt32.Parse (f.ToString ().Substring (0, 4));
               }
               catch (System.Exception e) {}
            }
            return 0;
         }
         set
         {
            SetTextFrame ("TDRC", value.ToString ());
         }
      }
      
      public override uint Track
      {
         get
         {
            try
            {
               return UInt32.Parse (GetFrames ("TRCK") [0].ToString ().Split (new char [] {'/'}) [0]);
            }
            catch (Exception e)
            {
               return 0;
            }
         }
         set
         {
            uint count = TrackCount;
            if (count != 0)
               SetTextFrame ("TRCK", value + "/" + count);
            else
               SetTextFrame ("TRCK", value.ToString ());
         }
      }
      
      public override uint TrackCount
      {
         get
         {
            try
            {
               return UInt32.Parse (GetFrames ("TRCK") [0].ToString ().Split (new char [] {'/'}) [1]);
            }
            catch (Exception e)
            {
               return 0;
            }
         }
         set
         {
            SetTextFrame ("TRCK", Track + "/" + value);
         }
      }
      
      public override uint Disc
      {
         get
         {
            try
            {
               return UInt32.Parse (GetFrames ("TPOS") [0].ToString ().Split (new char [] {'/'}) [0]);
            }
            catch (Exception e)
            {
               return 0;
            }
         }
         set
         {
            uint count = DiscCount;
            if (count != 0)
               SetTextFrame ("TPOS", value + "/" + count);
            else
               SetTextFrame ("TPOS", value.ToString ());
         }
      }
      
      public override uint DiscCount
      {
         get
         {
            try
            {
               return UInt32.Parse (GetFrames ("TPOS") [0].ToString ().Split (new char [] {'/'}) [1]);
            }
            catch (Exception e)
            {
               return 0;
            }
         }
         set
         {
            SetTextFrame ("TPOS", Disc + "/" + value);
         }
      }
      
      public override bool IsEmpty {get {return frame_list.Count == 0;}}

      public Header         Header         {get {return header;}}
      public ExtendedHeader ExtendedHeader {get {return extended_header;}}
      public Footer         Footer         {get {return footer;}}
      
      
      //////////////////////////////////////////////////////////////////////////
      // protected methods
      //////////////////////////////////////////////////////////////////////////
      protected void Read (TagLib.File file)
      {
         if (file == null)
            return;
            
         try {file.Mode = File.AccessMode.Read;}
         catch (Exception e) {return;}
         
         file.Seek (tag_offset);
         header.SetData (file.ReadBlock ((int) Header.Size));

         // if the tag size is 0, then this is an invalid tag (tags must contain
         // at least one frame)

         if(header.TagSize == 0)
            return;

         Parse (file.ReadBlock ((int) header.TagSize));
      }
      
      protected void Parse (ByteVector data)
      {
         int frame_data_position = 0;
         int frame_data_length = data.Count;

         // check for extended header

         if (header.ExtendedHeader)
         {
            if (ExtendedHeader == null)
               extended_header = new ExtendedHeader ();
            
            ExtendedHeader.SetData (data);
            
            if(ExtendedHeader.Size <= data.Count)
            {
               frame_data_position += (int) ExtendedHeader.Size;
               frame_data_length -= (int) ExtendedHeader.Size;
            }
         }

         // check for footer -- we don't actually need to parse it, as it *must*
         // contain the same data as the header, but we do need to account for its
         // size.

         if(header.FooterPresent && Footer.Size <= frame_data_length)
            frame_data_length -= (int) Footer.Size;

         // parse frames

         // Make sure that there is at least enough room in the remaining frame data for
         // a frame header.
         
         while (frame_data_position < frame_data_length - FrameHeader.Size (header.MajorVersion))
         {

            // If the next data is position is 0, assume that we've hit the padding
            // portion of the frame data.
            if(data [frame_data_position] == 0)
            {
               if (header.FooterPresent)
                  Debugger.Debug ("Padding *and* a footer found.  This is not allowed by the spec.");

               return;
            }

            Frame frame = factory.CreateFrame (data.Mid (frame_data_position), header.MajorVersion);

            if(frame == null)
               return;

            // Checks to make sure that frame parsed correctly.
            if(frame.Size < 0)
               return;

            frame_data_position += (int) (frame.Size + FrameHeader.Size (header.MajorVersion));            
            // Only add frames with content so we don't send out just we got in.
            if (frame.Size > 0)
               AddFrame (frame);
         }
      }
      
      protected void SetTextFrame (ByteVector id, string value)
      {
         if (value == null || value == "")
         {
            RemoveFrames (id);
            return;
         }
         
         SetTextFrame (id, new StringList (value));
      }
      
      protected void SetTextFrame (ByteVector id, StringList value)
      {
         if (value == null || value.Count == 0)
         {
            RemoveFrames (id);
            return;
         }
         
         Frame [] frames = GetFrames (id);
         if (frames.Length != 0)
            ((TextIdentificationFrame) frames [0]).SetText (value);
         else
         {
            TextIdentificationFrame f = new TextIdentificationFrame (id, factory.DefaultTextEncoding);
            AddFrame (f);
            f.SetText (value);
         }
      }
   }
}
