/***************************************************************************
    copyright            : (C) 2005 by Brian Nickel
    email                : brian.nickel@gmail.com
    based on             : mpcfile.cpp from TagLib
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

using System.Collections;
using System;

namespace TagLib.Mpc
{
   public enum TagType
   {
      NoTags  = 0x0000, // Empty set.  Matches no tag types.
      Id3v1   = 0x0001, // Matches ID3v1 tags.
      Id3v2   = 0x0002, // Matches ID3v2 tags.
      Ape     = 0x0004, // Matches APE tags.
      AllTags = 0xffff  // Matches all tag types.
   }

   public class File : TagLib.File
   {
      //////////////////////////////////////////////////////////////////////////
      // private properties
      //////////////////////////////////////////////////////////////////////////
      private Ape.Tag ape_tag;
      private long ape_location;
      private uint ape_size;

      private Id3v1.Tag id3v1_tag;
      private long id3v1_location;

      private Id3v2.Header id3v2_header;
      private long id3v2_location;
      private uint id3v2_size;

      private Tag tag;

      private Properties properties;

      // These indicate whether the file *on disk* has these tags, not if
      // this data structure does.  This is used in computing offsets.

      private bool has_ape;
      private bool has_id3v1;
      private bool has_id3v2;
      
      
      //////////////////////////////////////////////////////////////////////////
      // public methods
      //////////////////////////////////////////////////////////////////////////
      public File (string file, bool read_properties, Properties.ReadStyle properties_style) : base (file)
      {
         ape_tag        = null;
         ape_location   = -1;
         ape_size       = 0;
         id3v1_tag      = null;
         id3v1_location = -1;
         id3v2_header   = null;
         id3v2_location = -1;
         id3v2_size     = 0;
         tag            = null;
         properties     = null;
         has_ape        = false;
         has_id3v1      = false;
         has_id3v2      = false;
         
         try {Mode = AccessMode.Read;}
         catch (Exception e) {return;}
         
         Read (read_properties, properties_style);
         
         Mode = AccessMode.Closed;
      }

      public File (string file, bool read_properties) : this (file, read_properties, Properties.ReadStyle.Average)
      {
      }

      public File (string file) : this (file, true)
      {
      }
      
      public override bool Save ()
      {
         if (IsReadOnly)
         {
            Debugger.Debug ("Mpc.File.Save() -- File is read only.");
            return false;
         }
         
         try {Mode = AccessMode.Write;}
         catch (Exception e)
         {
            Debugger.Debug ("Mpc.File.Save() -- File is read only.");
            return false;
         }


         // Possibly strip ID3v2 tag

         if(has_id3v2 && id3v2_header == null)
         {
            RemoveBlock (id3v2_location, id3v2_size);
            has_id3v2 = false;
            if(has_id3v1)
            id3v1_location -= id3v2_size;
            if(has_ape)
            ape_location -= id3v2_size;
         }

         // Update ID3v1 tag

         if (id3v1_tag != null)
         {
            if(has_id3v1)
            {
               Seek (id3v1_location);
               WriteBlock (id3v1_tag.Render ());
            }
            else
            {
               Seek (0, System.IO.SeekOrigin.End);
               id3v1_location = Tell;
               WriteBlock (id3v1_tag.Render ());
               has_id3v1 = true;
            }
         }
         else if(has_id3v1)
         {
            RemoveBlock (id3v1_location, 128);
            has_id3v1 = false;
            if(has_ape && ape_location > id3v1_location)
                  ape_location -= 128;
         }

         // Update APE tag

         if (ape_tag != null)
         {
            if(has_ape)
               Insert (ape_tag.Render (), ape_location, ape_size);
            else
            {
               if (has_id3v1)
               {
                  Insert (ape_tag.Render (), id3v1_location, 0);
                  ape_size = ape_tag.Footer.CompleteTagSize;
                  has_ape = true;
                  ape_location = id3v1_location;
                  id3v1_location += ape_size;
               }
               else
               {
                  Seek (0, System.IO.SeekOrigin.End);
                  ape_location = Tell;
                  WriteBlock (ape_tag.Render ());
                  ape_size = ape_tag.Footer.CompleteTagSize;
                  has_ape = true;
               }
            }
         }
         else if (has_ape)
         {
            RemoveBlock (ape_location, ape_size);
            has_ape = false;
            if(has_id3v1 && id3v1_location > ape_location)
               id3v1_location -= ape_size;
         }

         Mode = AccessMode.Closed;
         return true;
      }
      
      public Id3v1.Tag GetId3v1Tag (bool create)
      {
         if (create && id3v1_tag == null)
         {
            // no ID3v1 tag exists and we've been asked to create one
            id3v1_tag = new Id3v1.Tag ();
            tag = new CombinedTag (ape_tag, id3v1_tag);
         }
         
         return id3v1_tag;
      }
      
      public Id3v1.Tag GetId3v1Tag ()
      {
         return GetId3v1Tag (false);
      }
      
      public Ape.Tag GetApeTag (bool create)
      {
         if (create && ape_tag == null)
         {
            // no APE tag exists and we've been asked to create one
            ape_tag = new Ape.Tag ();
            tag = new CombinedTag(ape_tag, id3v1_tag);
         }
         
         return ape_tag;
      }
      
      public Ape.Tag GetApeTag ()
      {
         return GetApeTag (false);
      }
      
      public void Remove (int tags)
      {
         if ((tags & (int) TagType.Id3v1) != 0)
            id3v1_tag = null;

         if ((tags & (int) TagType.Id3v2) != 0)
            id3v2_header = null;

         if ((tags & (int) TagType.Ape) != 0)
            ape_tag = null;

         tag = new CombinedTag(ape_tag, id3v1_tag);
      }
      
      public void Remove ()
      {
         Remove ((int) TagType.AllTags);
      }
      
      
      //////////////////////////////////////////////////////////////////////////
      // public properties
      //////////////////////////////////////////////////////////////////////////
      public override TagLib.Tag Tag {get {return tag;}}
      
      public override TagLib.AudioProperties AudioProperties {get {return properties;}}
      
      
      //////////////////////////////////////////////////////////////////////////
      // private methods
      //////////////////////////////////////////////////////////////////////////
      private void Read (bool read_properties, Properties.ReadStyle properties_style)
      {
         // Look for an ID3v1 tag

         id3v1_location = FindId3v1 ();

         if (id3v1_location >= 0)
         {
            id3v1_tag = new Id3v1.Tag (this, id3v1_location);
            has_id3v1 = true;
         }

         // Look for an APE tag

         ape_location = FindApe ();

         if (ape_location >= 0)
         {
            ape_tag = new Ape.Tag (this, ape_location);
            ape_size = ape_tag.Footer.CompleteTagSize;
            ape_location = ape_location + Ape.Footer.Size - ape_size;
            has_ape = true;
         }

         tag = new CombinedTag (ape_tag, id3v1_tag);

         // Look for and skip an ID3v2 tag

         id3v2_location = FindId3v2 ();

         if(id3v2_location >= 0)
         {
            Seek (id3v2_location);
            id3v2_header = new Id3v2.Header (ReadBlock ((int) Id3v2.Header.Size));
            id3v2_size = id3v2_header.CompleteTagSize;
            has_id3v2 = true;
         }

         if(has_id3v2)
            Seek (id3v2_location + id3v2_size);
         else
            Seek (0);

         // Look for MPC metadata

         if (read_properties)
            properties = new Properties (ReadBlock ((int) Properties.HeaderSize), Length - id3v2_size - ape_size);
      }
      
      private long FindApe ()
      {
         if (!IsValid)
            return -1;

         if(has_id3v1)
            Seek (-160, System.IO.SeekOrigin.End);
         else
            Seek (-32, System.IO.SeekOrigin.End);

         long p = Tell;

         if(ReadBlock (8) == Ape.Tag.FileIdentifier)
            return p;

         return -1;
      }
      
      private long FindId3v1 ()
      {
         if (!IsValid)
            return -1;

         Seek (-128, System.IO.SeekOrigin.End);
         long p = Tell;

         if(ReadBlock (3) == Id3v1.Tag.FileIdentifier)
            return p;

         return -1;
      }
      
      private long FindId3v2 ()
      {
         if (!IsValid)
            return -1;

         Seek (0);

         if(ReadBlock (3) == Id3v2.Header.FileIdentifier)
            return 0;

         return -1;
      }
   }
}
