/***************************************************************************
    copyright            : (C) 2005 by Brian Nickel
    email                : brian.nickel@gmail.com
    based on             : mpegfile.cpp from TagLib
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

using System.Collections;
using System;

namespace TagLib.Mpeg
{
   public enum TagType
   {
      NoTags  = 0x0000, // Empty set.  Matches no tag types.
      Id3v1   = 0x0001, // Matches ID3v1 tags.
      Id3v2   = 0x0002, // Matches ID3v2 tags.
      Ape     = 0x0004, // Matches APE tags.
      AllTags = 0xffff  // Matches all tag types.
   }

   public class File : TagLib.File
   {
      //////////////////////////////////////////////////////////////////////////
      // private properties
      //////////////////////////////////////////////////////////////////////////
      private Id3v2.FrameFactory id3v2_frame_factory;
      private Id3v2.Tag          id3v2_tag;
      private long               id3v2_location;
      private uint               id3v2_original_size;

      private Ape.Tag            ape_tag;
      private long               ape_location;
      private uint               ape_original_size;

      private Id3v1.Tag          id3v1_tag;
      private long               id3v1_location;

      private Tag                tag;

      // These indicate whether the file *on disk* has these tags, not if
      // this data structure does.  This is used in computing offsets.

      private bool               has_id3v2;
      private bool               has_id3v1;
      private bool               has_ape;

      private Properties         properties;
      
      
      //////////////////////////////////////////////////////////////////////////
      // public methods
      //////////////////////////////////////////////////////////////////////////
      public File (string file, Id3v2.FrameFactory frame_factory, bool read_properties, Properties.ReadStyle properties_style) : base (file)
      {
         id3v2_frame_factory = frame_factory;
         id3v2_tag           = null;
         id3v2_location      = -1;
         id3v2_original_size = 0;
         ape_tag             = null;
         ape_location        = -1;
         ape_original_size   = 0;
         id3v1_tag           = null;
         id3v1_location      = -1;
         tag                 = null;
         has_id3v2           = false;
         has_id3v1           = false;
         has_ape             = false;
         properties          = null;
         
         try {Mode = AccessMode.Read;}
         catch (Exception e) {return;}
         
         tag = new Tag (this);
         Read (read_properties, properties_style);
         
         Mode = AccessMode.Closed;
      }

      public File (string file, Id3v2.FrameFactory frame_factory, bool read_properties) : this (file, frame_factory, read_properties, Properties.ReadStyle.Average)
      {
      }
      
      public File (string file, Id3v2.FrameFactory frame_factory) : this (file, frame_factory, true)
      {
      }
      
      public File (string file, bool read_properties, Properties.ReadStyle properties_style) : this (file, Id3v2.FrameFactory.Instance, read_properties, properties_style)
      {
      }
      
      public File (string file, bool read_properties) : this (file, read_properties, Properties.ReadStyle.Average)
      {
      }
      
      public File (string file) : this (file, true)
      {
      }

      public bool Save (int tags, bool strip_others)
      {
         if(tags == (int) TagType.NoTags && strip_others)
            return Strip ((int) TagType.AllTags);

         if (id3v2_tag == null && id3v1_tag == null && ape_tag == null)
         {
            if ((has_id3v1 || has_id3v2 || has_ape) && strip_others)
            return Strip ((int) TagType.AllTags);

            return true;
         }

         if (IsReadOnly)
         {
            Debugger.Debug ("Mpeg.File.Save() -- File is read only.");
            return false;
         }
         
         try {Mode = AccessMode.Write;}
         catch (Exception e)
         {
            Debugger.Debug ("Mpeg.File.Save() -- File is read only.");
            return false;
         }

         // Create the tags if we've been asked to.  Copy the values from the tag that
         // does exist into the new tag.

         if ((tags & (int) TagType.Id3v2) != 0 && id3v1_tag != null)
            Tag.Duplicate (id3v1_tag, GetId3v2Tag (true), false);

         if ((tags & (int) TagType.Id3v1) != 0 && id3v2_tag != null)
            Tag.Duplicate (id3v2_tag, GetId3v1Tag (true), false);

         bool success = true;

         if(((int) TagType.Id3v2 & tags) != 0)
         {
            if(id3v2_tag != null && !id3v2_tag.IsEmpty)
            {
               if (!has_id3v2)
                  id3v2_location = 0;

               Insert (id3v2_tag.Render (), id3v2_location, id3v2_original_size);
            }
            else if(strip_others)
               success = Strip ((int) TagType.Id3v2, false) && success;
         }
         else if (has_id3v2 && strip_others)
            success = Strip ((int) TagType.Id3v2) && success;

         if(((int) TagType.Id3v1 & tags) != 0)
         {
            if(id3v1_tag != null && !id3v1_tag.IsEmpty)
            {
               int offset = has_id3v1 ? -128 : 0;
               Seek (offset, System.IO.SeekOrigin.End);
               id3v1_location = Tell;
               has_id3v1 = true;
               WriteBlock (id3v1_tag.Render ());         
            }
            else if(strip_others)
               success = Strip ((int) TagType.Id3v1) && success;
         }
         else if (has_id3v1 && strip_others)
            success = Strip ((int) TagType.Id3v1, false) && success;

         // Dont save an APE-tag unless one has been created
         if(((int) TagType.Ape & tags) != 0 && ape_tag != null)
         {
            if (has_ape)
               Insert (ape_tag.Render (), ape_location, ape_original_size);
            else
            {
               if(has_id3v1)
               {
                  Insert (ape_tag.Render (), id3v1_location, 0);
                  ape_original_size = ape_tag.Footer.CompleteTagSize;
                  has_ape = true;
                  ape_location = id3v1_location;
                  id3v1_location += ape_original_size;
               }
               else
               {
                  Seek (0, System.IO.SeekOrigin.End);
                  ape_location = Tell;
                  WriteBlock (ape_tag.Render ());
                  ape_original_size = ape_tag.Footer.CompleteTagSize;
                  has_ape = true;
               }
            }
         }
         else if(has_ape && strip_others)
            success = Strip ((int) TagType.Ape, false) && success;

         Mode = AccessMode.Closed;
         return success;
      }

      public bool Save (int tags)
      {
         return Save (tags, true);
      }

      public override bool Save ()
      {
         return Save ((int) TagType.AllTags);
      }


      public Id3v2.Tag GetId3v2Tag (bool create)
      {
         if (create && id3v2_tag == null)
            id3v2_tag = new Id3v2.Tag ();

         return id3v2_tag;
      }

      public Id3v2.Tag GetId3v2Tag ()
      {
         return GetId3v2Tag (false);
      }

      public Id3v1.Tag GetId3v1Tag (bool create)
      {
         if (create && id3v1_tag == null)
            id3v1_tag = new Id3v1.Tag ();

         return id3v1_tag;
      }

      public Id3v1.Tag GetId3v1Tag ()
      {
         return GetId3v1Tag (false);
      }

      public Ape.Tag GetApeTag (bool create)
      {
         if (create && ape_tag == null)
            ape_tag = new Ape.Tag ();

         return ape_tag;
      }

      public Ape.Tag GetApeTag ()
      {
         return GetApeTag (false);
      }

      public bool Strip (int tags, bool free_memory)
      {
         AccessMode original_mode = Mode;
         
         if (IsReadOnly)
         {
            Debugger.Debug ("Mpeg.File.Strip() - Cannot strip tags from a read only file.");
            return false;
         }
         
         try {Mode = AccessMode.Write;}
         catch (Exception e)
         {
            Debugger.Debug ("Mpeg.File.Strip() - Cannot strip tags from a read only file.");
            return false;
         }

         if ((tags & (int) TagType.Id3v2) != 0 && has_id3v2)
         {
            RemoveBlock (id3v2_location, id3v2_original_size);
            id3v2_location = -1;
            id3v2_original_size = 0;
            has_id3v2 = false;
            
            if (free_memory)
               id3v2_tag = null;

            // v1 tag location has changed, update if it exists
            if (id3v1_tag != null)
               id3v1_location = FindId3v1 ();
         }

         if ((tags & (int) TagType.Id3v1) != 0 && has_id3v1)
         {
            Truncate (id3v1_location);
            id3v1_location = -1;
            has_id3v1 = false;
            
            if (free_memory)
               id3v1_tag = null;
         }

         if ((tags & (int) TagType.Ape) != 0 && has_ape)
         {
            RemoveBlock (ape_location, ape_original_size);
            ape_location = -1;
            has_ape = false;
            
            if (has_id3v1)
               id3v1_location = FindId3v1 ();
            
            if (free_memory)
               ape_tag = null;
         }

         Mode = original_mode;
         return true;
      }

      public bool Strip (int tags)
      {
         return Strip (tags, true);
      }

      public bool Strip ()
      {
         return Strip ((int) TagType.AllTags);
      }
      
      public long NextFrameOffset (long position)
      {
         ByteVector buffer = (byte) 0;
         bool previous_partial_synch_match = false;
         
         while (buffer.Count > 0)
         {
            Seek (position);
            buffer = ReadBlock ((int) BufferSize);

            if (previous_partial_synch_match && SecondSynchByte (buffer [0]))
               return position - 1;
            
            for (int i = 0; i < buffer.Count - 1; i++)
               if (buffer [i] == 0xFF && SecondSynchByte (buffer [i + 1]))
                  return position + i;
            
            if (buffer [buffer.Count - 1] == 0xFF)
               previous_partial_synch_match = true;
            
            position += BufferSize;
         }
         
         return -1;
      }

      public long PreviousFrameOffset (long position)
      {
         bool previous_partial_synch_match = false;
         
         while (position - BufferSize > BufferSize)
         {
            position -= BufferSize;
            Seek (position);
            ByteVector buffer = ReadBlock ((int) BufferSize);

            if (previous_partial_synch_match && SecondSynchByte (buffer [(int) BufferSize - 1]))
               return position + BufferSize - 1;
            
            // If the amount of data is smaller than an MPEG header (4 bytes) there's no
            // chance of this being valid.

            if (buffer.Count < 4)
               return -1;

            for (int i = buffer.Count - 2; i >= 0; i--)
               if (buffer [i] == 0xFF && SecondSynchByte (buffer [i + 1]))
                  return position + i;
            
            if (buffer [0] == 0xFF)
               previous_partial_synch_match = true;
         }

         return -1;
      }
      
      
      //////////////////////////////////////////////////////////////////////////
      // public properties
      //////////////////////////////////////////////////////////////////////////
      public long FirstFrameOffset
      {
         get
         {
            return NextFrameOffset (has_id3v2 ? id3v2_location + id3v2_tag.Header.CompleteTagSize : 0);
         }
      }

      public long LastFrameOffset
      {
         get
         {
            return PreviousFrameOffset (has_id3v1 ? id3v1_location - 1 : Length);
         }
      }

      public Id3v2.FrameFactory Id3v2FrameFactory
      {
         set {id3v2_frame_factory = value;}
      }

      public override TagLib.Tag Tag {get {return tag;}}
      
      public override AudioProperties AudioProperties {get {return properties;}}
      
      
      //////////////////////////////////////////////////////////////////////////
      // private methods
      //////////////////////////////////////////////////////////////////////////
      private void Read (bool read_properties, Properties.ReadStyle properties_style)
      {
         // Look for an ID3v2 tag
         id3v2_location = FindId3v2 ();

         if (id3v2_location >= 0)
         {
            id3v2_tag = new Id3v2.Tag (this, id3v2_location, id3v2_frame_factory);

            id3v2_original_size = id3v2_tag.Header.CompleteTagSize;

            if (id3v2_tag.Header.TagSize <= 0)
               id3v2_tag = null;
            else
               has_id3v2 = true;
         }

         // Look for an ID3v1 tag
         id3v1_location = FindId3v1 ();

         if (id3v1_location >= 0)
         {
            id3v1_tag = new Id3v1.Tag (this, id3v1_location);
            has_id3v1 = true;
         }

         // Look for an APE tag
         ape_location = FindApe ();

         if (ape_location >= 0)
         {
            ape_tag = new Ape.Tag (this, ape_location);
            ape_original_size = ape_tag.Footer.CompleteTagSize;
            ape_location = ape_location + Ape.Footer.Size - ape_original_size;

            has_ape = true;
         }

         if(read_properties)
            properties = new Properties (this, properties_style);
      }

      private long FindId3v2 ()
      {
         // This method is based on the contents of TagLib::File::find(), but because
         // of some subtlteies -- specifically the need to look for the bit pattern of
         // an MPEG sync, it has been modified for use here.
         
         if (IsValid && Id3v2.Header.FileIdentifier.Count <= BufferSize)
         {
            // The position in the file that the current buffer starts at.

            long buffer_offset = 0;
            ByteVector buffer;

            // These variables are used to keep track of a partial match that happens at
            // the end of a buffer.

            int previous_partial_match = -1;
            bool previous_partial_synch_match = false;

            // Save the location of the current read pointer.  We will restore the
            // position using seek() before all returns.

            long original_position = Tell;

            // Start the search at the beginning of the file.

            Seek (0);

            // This loop is the crux of the find method.  There are three cases that we
            // want to account for:
            // (1) The previously searched buffer contained a partial match of the search
            // pattern and we want to see if the next one starts with the remainder of
            // that pattern.
            //
            // (2) The search pattern is wholly contained within the current buffer.
            //
            // (3) The current buffer ends with a partial match of the pattern.  We will
            // note this for use in the next itteration, where we will check for the rest
            // of the pattern.

            for (buffer = ReadBlock ((int) BufferSize); buffer.Count > 0; buffer = ReadBlock ((int) BufferSize))
            {

               // (1) previous partial match

               if(previous_partial_synch_match && SecondSynchByte (buffer [0]))
                  return -1;

               if (previous_partial_match >= 0 && (int) BufferSize > previous_partial_match)
               {
                  int pattern_offset = (int) (BufferSize - previous_partial_match);
                  if (buffer.ContainsAt (Id3v2.Header.FileIdentifier, 0, pattern_offset))
                  {
                     Seek (original_position);
                     return buffer_offset - BufferSize + previous_partial_match;
                  }
               }

               // (2) pattern contained in current buffer

               long location = buffer.Find (Id3v2.Header.FileIdentifier);
               if (location >= 0)
               {
                  Seek (original_position);
                  return buffer_offset + location;
               }

               int first_synch_byte = buffer.Find ((byte) 0xFF);

               // Here we have to loop because there could be several of the first
               // (11111111) byte, and we want to check all such instances until we find
               // a full match (11111111 111) or hit the end of the buffer.

               while (first_synch_byte >= 0)
               {
                  // if this *is not* at the end of the buffer
                  
                  if (first_synch_byte < (int) buffer.Count - 1)
                  {
                     if (SecondSynchByte (buffer [first_synch_byte + 1]))
                     {
                        // We've found the frame synch pattern.
                        Seek (original_position);
                        return -1;
                     }
                     else
                     {
                        // We found 11111111 at the end of the current buffer indicating a
                        // partial match of the synch pattern.  The find() below should
                        // return -1 and break out of the loop.

                        previous_partial_synch_match = true;
                     }
                  }

                  // Check in the rest of the buffer.

                  first_synch_byte = buffer.Find ((byte) 0xFF, first_synch_byte + 1);
               }

               // (3) partial match

               previous_partial_match = buffer.EndsWithPartialMatch (Id3v2.Header.FileIdentifier);

               buffer_offset += BufferSize;
            }

            Seek (original_position);
         }
         
         return -1;
      }

      private long FindId3v1 ()
      {
         if (IsValid)
         {
            Seek (-128, System.IO.SeekOrigin.End);
            long p = Tell;

            if (ReadBlock (3) == Id3v1.Tag.FileIdentifier)
            return p;
         }
         
         return -1;
      }

      private long FindApe ()
      {
         if (IsValid)
         {
            if (has_id3v1)
               Seek (-160, System.IO.SeekOrigin.End);
            else
               Seek (-32, System.IO.SeekOrigin.End);

            long p = Tell;

            if (ReadBlock (8) == Ape.Tag.FileIdentifier)
               return p;
         }
         return -1;
      }
      
      private bool SecondSynchByte (byte b)
      {
         return b >= 0xE0;
      }
   }
}
